"""
Copyright (c) Jordan Maxwell, All Rights Reserved.
See LICENSE file in the project root for full license information.

This module contains classes and functions for interacting with droid scripts.

1. DroidScripts: An enumeration containing constants representing available droid scripts.
2. DroidScriptActions: An enumeration containing constants representing available droid script actions.
3. DroidScriptEngine: A class that represents the droid script engine and provides methods for executing droid scripts.
"""

import asyncio
import logging
from datetime import datetime
from dbeacon import scanner, beacon
from droiddepot.protocol import DroidCommandId

class DroidScripts(object):
    """
    An enumeration containing constants representing available droid scripts.
    """

    GeneralParkResponseScript = 1
    DroidDepotParkResponseScript = 2
    ResistenceParkResponseScript = 3
    UnknownParkResponseScript = 4 # Possibly related to smuggers run?
    OgasCantinaParkResponseScript = 5
    DokOndarsParkResponseScript = 6
    FirstOrderParkResponseScript = 7
    DroidBayActivationSequence = 8
    UnknownBUnitScript9 = 9
    UnknownBUnitScript10 = 10
    DroidPairingSequence1 = 11
    DroidPairingSequence2 = 12
    FullThrottleTestScript = 13

class DroidScriptActions(object):
    """
    An enumeration containing constants representing available droid script actions.
    """

    OpenScript = 0
    CloseScript = 1
    ExecuteScript = 2

class DroidScriptEngine(object):
    """
    A class that represents the droid script engine and provides methods for executing droid scripts.
    """

    def __init__(self, droid: object) -> None:
        """
        Initializes a new instance of the DroidScriptEngine class.

        Args:
            droid (object): The droid instance to use.
        """

        self.droid = droid

        self.__location_reaction_tracker = {}
        self.reaction_scanner = scanner.DBeaconScanner()
        self.reaction_scanner.add_beacon_handler(10, self.__perform_location_reactions)

    async def send_script_command(self, script_id: int, script_action: int) -> None:
        """
            Sends a script command to the droid.

            Args:
                script_id (int): The ID of the script to execute.
                script_action (int): The action to perform with the script.

            Raises:
                ValueError: If the script ID is 13 (FullThrottleTestScript) or
                the script is less then 0.
        """

        if script_id <= 0 and script_action != DroidScriptActions.CloseScript:
            raise ValueError("Invalid script id requested. Script ids must be larger then 0")

        if script_id == 13:
            raise ValueError("Attempted to use a dangerous script. Execution denied")

        command_data = "%s%s" % ("{:02d}".format(script_id), "{:02d}".format(script_action))
        await self.droid.send_droid_command(DroidCommandId.ScriptActionComand, command_data)

    async def execute_script(self, script_id: int) -> None:
        """
        Executes a droid script.

        Args:
            script_id (int): The ID of the script to execute.
        """

        await self.send_script_command(script_id, DroidScriptActions.ExecuteScript)

    async def execute_location_beacon(self, beacon: beacon.LocationBeacon) -> None:
        """
        Executes a location beacon on the connected droid emulation what would happen
        if the droid encountered the beacon at a Disney park

        Args:
            beacon (LocationBeacon): Location beacon to execute
        """

        await self.execute_script(beacon.location_id)

    async def execute_location_reaction(self, location_id: int) -> None:
        """
        Executes a location reaction on the connected droid emulation what would happen
        
        Args:
            location_id (int): Location id to execute
        """

        if location_id < 0 or location_id > 7:
            raise ValueError("Invalid location id requested. Location ids must be between 0 and 7")
        
        await self.execute_script(location_id)

    async def open_script(self, script_id: int) -> None:
        """
        Opens a droid script for writing. All command sent after the script is opened will be
        written into the script id until close_script is called

        Args:
            script_id (int): The ID of the script to open for writing.
        """

        if script_id <= 0:
            raise ValueError('Script ids must be larger then 0')

        if script_id >= 1 and script_id <= 13:
            raise ValueError("Attempted to rewrite Disney programmed scripts. Action prevented for safety")

        await self.send_script_command(script_id, DroidScriptActions.OpenScript)

    async def close_script(self, script_id: int = 0) -> None:
        """
        Closes the currently open script. This will also stop any script currently
        being executed.

        Args:
            script_id (int): The ID of the script to close.
        """

        await self.send_script_command(script_id, DroidScriptActions.CloseScript)

    def __calculate_reaction_time(self, interval: int) -> int:
        """
        Calculates the time in seconds from a park beacon's reaction interval.

        Args:
            interval (int): Reaction interval received from a location beacon
        """

        interval = interval * 5
        if interval < 60:
            interval = 60

        return interval

    async def __perform_location_reactions(self, beacons: list) -> None:
        """
        Executes a script associated with each park location beacon that the droid enters.

        Args:
            locations (list): A list of beacons detected

        Returns:
            int: The calculated reaction time in seconds.
        """

        # Verify we have at least one location to react to first.
        if len(beacons) == 0:
            return
        
        can_execute = True
        already_executed = False
        location_beacon_address = "Unknown"

        for location_beacon_info in beacons:
            try:
                location_beacon_address, location_beacon = location_beacon_info

                # Check if we already reacted and if we have check if we are in a new reaction window
                if location_beacon_address in self.__location_reaction_tracker:
                    last_execution = self.__location_reaction_tracker[location_beacon_address]
                    time_since_last = (datetime.now() - last_execution).total_seconds()
                    can_execute = time_since_last >= self.__calculate_reaction_time(location_beacon.reaction_interval)

                # Attempt to execute the reaction
                if can_execute and already_executed == False:
                    await self.execute_location_reaction(location_beacon.location_id)
                    self.__location_reaction_tracker[location_beacon_address] = datetime.now()
                    already_executed = True
            except Exception as e:
                logging.error('An unexpected error occured processing a park location beacon: %s' % location_beacon_address)
                logging.error(e, exc_info=True) 

        if already_executed:
            await asyncio.sleep(5)

    def start_beacon_reactions(self) -> None:
        """
        Enables SWGE park beacon reactions similar to the internal firmware.
        """

        self.reaction_scanner.start()

    def stop_beacon_reactions(self) -> None:
        """
        Disables park beacon reactions.
        """

        self.reaction_scanner.stop()


"""
Copyright (c) Jordan Maxwell, All Rights Reserved.
See LICENSE file in the project root for full license information.

This module provides classes for processing and handling notifications from a connected droid. 

It contains the DroidNotifyMessage class, which represents a notification message from the droid, and 
the DroidNotificationProcessor class, which handles incoming notifications from the droid and passes 
them to the appropriate handlers. The DroidNotificationProcessor class also includes methods for decoding 
incoming messages and verifying firmware versions.
"""

import logging
import asyncio
from droiddepot.utils import hex_to_int
from droiddepot.hardware import DroidFirmwareVersion
from droiddepot.protocol import *

class DroidNotifyMessage(object):
    """
    Represents a notification message received from a droid.

    Attributes:
        message_size (int): The size of the message.
        unknown1 (int): An unknown integer value.
        command_id (int): The ID of the command associated with the notification.
        unknown3 (int): Another unknown integer value.
        message_data (str): The data associated with the notification.
    """

    def __init__(self, message_size: int, unknown1: int, command_id: int, unknown3: int, message_data: str):
        """
        Initializes a new instance of the DroidNotifyMessage class.

        Args:
            message_size (int): The size of the message.
            unknown1 (int): An unknown integer value.
            command_id (int): The ID of the command associated with the notification.
            unknown3 (int): Another unknown integer value.
            message_data (str): The data associated with the notification.
        """

        self.message_size = message_size
        self.unknown1 = unknown1
        self.command_id = command_id
        self.unknown3 = unknown3
        self.message_data = message_data

    def __str__(self) -> str:
        """
        Returns a string representation of the DroidNotifyMessage instance.

        Returns:
            str: A string representation of the DroidNotifyMessage instance.
        """

        return ('Size: %s, Unknown1: %s, Command Id: %s, Unknown2: %s, Data: %s' % (
            self.message_size, self.unknown1, self.command_id, self.unknown3, self.message_data))

class DroidNotificationProcessor(object):
    """
    Processes notification events from a connected droid.

    Attributes:
        droid (object): The connected droid instance.
    """

    def __init__(self, droid: object) -> None:
        """
        Initializes a new instance of the DroidNotificationProcessor class.

        Args:
            droid (object): The connected droid instance.
        """

        self.droid = droid
        self.__pending_callback_events = {}

    async def wait_for_command_response(self, command_id: int, timeout = 1.0) -> object:
        """
        Waits for a command response from the droid's microcontroller.

        Args:
            command_id (int): The command id to expect as a response
            timeout (float): The maximum time to wait for a response, in seconds.

        Raises:
            TimeoutError: If a response is not received within the specified timeout.

        Returns:
            An object representing the response to the given command, or None if no response is received within the specified timeout.
        """

        # Queue our callback Queue for handling
        callback_queue = asyncio.Queue()
        if command_id not in self.__pending_callback_events:
            self.__pending_callback_events[command_id] = []
        self.__pending_callback_events[command_id].append(callback_queue)

        # Wait for our response from the droid
        try:
            return await asyncio.wait_for(callback_queue.get(), timeout=timeout)
        except asyncio.TimeoutError:
            return None

    def decode_notify_message(self, data: bytearray) -> DroidNotifyMessage:
        """
        Decodes an incoming message to a DroidNotifyMessage instance.

        Args:
            data (bytearray): The incoming message data.

        Returns:
            DroidNotifyMessage: A DroidNotifyMessage instance representing the incoming message.
        """

        hex_data = data.hex()
        message_size = hex_to_int(hex_data[:2]) - 0x1f
        unknown1 = hex_to_int(hex_data[2:4])
        command_id = hex_to_int(hex_data[4:6])
        unknown3 = hex_to_int(hex_data[6:8])
        message_data = hex_data[8:]

        if len(data) != message_size:
            raise ValueError('Received truncated packet. Expected %s, got %s' % (len(data), message_size))

        return DroidNotifyMessage(message_size, unknown1, command_id, unknown3, message_data)

    async def handle_incoming_message(self, sender: object, data: bytearray) -> None:
        """
        Processes notification events from the connected droid.

        Args:
            sender (object): The sender of the notification event.
            data (bytearray): The notification event data.
        """

        try:
            message = self.decode_notify_message(data)
            await self.__process_incoming_message(message)
        except ValueError as e:
            logging.error('Failed to process notification message with data %s. Data is malformed' % (data.hex()))
            logging.error(e, exc_info=True)

    async def __handle_pending_callbacks(self, message: DroidNotifyMessage, response: object) -> None:
        """
        Handles any pending callbacks registered for the given command ID.

        Args:
            message (DroidNotifyMessage): The notification message received from the droid.
            response (object): The response object to be passed to the callbacks.
        """

        if message.command_id not in self.__pending_callback_events:
            return
        
        command_id = message.command_id
        if not len(self.__pending_callback_events[command_id]):
            return
        
        first_callback_event = self.__pending_callback_events[command_id][0]
        await first_callback_event.put(response)
        self.__pending_callback_events[command_id].remove(first_callback_event)

    async def __process_incoming_message(self, message: DroidNotifyMessage) -> None:
        """
        Processes the parsed command message from the connected droid and passes it
        to the appropriate command handler.

        Args:
            message (DroidNotifyMessage): The parsed command message received from the droid.
        """

        if not DroidCommandId.valid_command(message.command_id):
            logging.warning('Received unknown command %s. Ignoring' % message.command_id)
            return
        
        response = None
        if message.command_id == DroidCommandId.RetrieveFirmwareInformationResponse:
            response = await self.__verify_firmware_version(message)
        elif message.command_id == DroidCommandId.RUnitHeadEvent:
            response = await self.__handle_runit_head_motor_events(message)
        else:
            logging.warning('No handler present for droid command: %s (%s)' % (DroidCommandId(message.command_id).name, message.message_data))

        if response == None:
            response = message.message_data
        await self.__handle_pending_callbacks(message, response)

    async def __verify_firmware_version(self, message: DroidNotifyMessage) -> None:
        """
        Verifies that the firmware version received from the connected droid matches the expected firmware version.

        Args:
            message (DroidNotifyMessage): The parsed command message received from the droid.

        Raises:
            Exception: If the firmware version received does not match the expected firmware version.
        """

        if message.message_data != DroidFirmwareVersion:
            raise Exception('Possibly incomaptible droid detected. Possibly a new firmware version.')
        
    async def __handle_runit_head_motor_events(self, message: DroidNotifyMessage) -> None:
        """
        Handles R unit head motor events received from the connected droid.

        Args:
            message (DroidNotifyMessage): The parsed command message received from the droid.
        """

        unknown1 = hex_to_int(message.message_data[:2])
        event_id = hex_to_int(message.message_data[2:4])

        await self.droid.motor_controller.process_runit_head_motor_event(event_id)


"""
Copyright (c) Jordan Maxwell, All Rights Reserved.
See LICENSE file in the project root for full license information.

DroidConnection is a BLE class representing a connection to a SWGE DroidDepot droid. 
It includes methods for connecting, disconnecting, sending commands, and running scripts on the droid.

It also includes instances of DroidAudioController, DroidMotorController, and DroidScriptEngine 
to manage the droid's audio, motor, and script functions.
"""

import asyncio
import logging
from time import sleep
from threading import Thread
from bleak import BleakScanner, BleakClient
from droiddepot.protocol import *
from droiddepot.audio import DroidAudioController
from droiddepot.motor import DroidMotorController
from droiddepot.script import DroidScriptEngine, DroidScripts
from droiddepot.voice import DroidVoiceController
from droiddepot.notify import DroidNotificationProcessor
from droiddepot.hardware import DroidPersonalityIdentifier, DroidAffiliation

class DroidConnection(object):
    """
    Represents a connection to a SWGE DroidDepot droid.

    Args:
        profile (str): A string representing the UUID of the BLE profile to connect to.
        manufacturer_data (dict): A dictionary containing the manufacturer data of the droid being connected.
    """

    DroidServiceId = '09b600a0-3e42-41fc-b474-e9c0c8f0c801'

    def __init__(self, profile: str, manufacturer_data):
        """
        Initializes a new instance of the Droid class.

        Args:
            profile (str): A string representing the UUID of the BLE profile to connect to.
            manufacturer_data (dict): A dictionary containing the manufacturer data of the droid being connected.

        Attributes:
            droid: A BLE connection object for the droid.
            personality_id: The personality ID of the droid. Default is DroidPersonalityIdentifier.RUnit.
            affiliation_id: The affiliation ID of the droid. Default is DroidAffiliation.Scoundrel.
            audio_controller: An instance of the DroidAudioController class.
            script_engine: An instance of the DroidScriptEngine class.
            motor_controller: An instance of the DroidMotorController class.
            heartbeat_loop: An asyncio event loop used for the heartbeat thread.
            heartbeat_thread: A thread that runs the heartbeat_loop.
        """
        
        self.profile = profile
        self.droid = None
        self.manufacturer_data = manufacturer_data
        self.personality_id = DroidPersonalityIdentifier.RUnit
        self.affiliation_id = DroidAffiliation.Scoundrel

        self.audio_controller = DroidAudioController(self)
        self.script_engine = DroidScriptEngine(self)
        self.motor_controller = DroidMotorController(self)
        self.voice_controller = DroidVoiceController(self)
        self.notify_processor = DroidNotificationProcessor(self)

        self.heartbeat_loop = asyncio.new_event_loop()
        self.heartbeat_thread = None

    async def connect(self, silent: bool = False) -> None:
        """
        Connect to the Droid using BLE.
        """

        timeout = 0.0
        self.droid = BleakClient(self.profile)
        await self.droid.connect()
        await self.droid.start_notify(DroidBluetoothCharacteristics.DroidNotifyCharacteristic, self.notification_handler)

        while not self.droid.is_connected and timeout < 10:
            sleep (.1)
            timeout += .1

        connect_code = bytearray.fromhex("222001")
        await self.droid.write_gatt_char(0x000d, connect_code, False)
        await self.droid.write_gatt_char(0x000d, connect_code, False)

        droid_data = self.manufacturer_data[DisneyBLEManufacturerId.DroidManufacturerId]

        droid_data_len = len(droid_data)
        self.personality_id = droid_data[droid_data_len - 1]
        self.affiliation_id = (droid_data[droid_data_len - 2] - 0x80) / 2
        
        if not silent:
            await self.script_engine.execute_script(DroidScripts.DroidPairingSequence1)
            sleep(4)

        self.heartbeat_thread = Thread(target=self.__start_heartbeat_loop, args=(self.heartbeat_loop,), daemon=True)
        self.heartbeat_thread.start()
        asyncio.run_coroutine_threadsafe(self.__send_heartbeat_command(), self.heartbeat_loop)

    async def notification_handler(self, sender: object, data: bytearray) -> None:
        """
        Processes notification events from the connected droid and
        passes it to our notify message processor.
        """

        await self.notify_processor.handle_incoming_message(sender, data)

    def __start_heartbeat_loop(self, loop: asyncio.AbstractEventLoop) -> None:
        """
        Starts the heartbeat event loop
        """

        asyncio.set_event_loop(loop)
        loop.run_forever()

    async def __send_heartbeat_command(self) -> None:
        """
        Sends a harmless unused command every 10 seconds to keep our connection to the droid alive even when not in use.
        """

        while self.droid.is_connected:
            await self.send_droid_command(DroidCommandId.ConnectionHeartbeat)
            sleep(10)

    async def disconnect(self, silent: bool = False) -> None:
        """
        Disconnect from the Droid.
        """

        if not self.droid.is_connected:
            return

        logging.info("Disconnecting from droiddepot")
        try:
            if not silent:
                await self.audio_controller.play_shutdown_audio()
        finally:
            await self.droid.disconnect()

            if self.heartbeat_loop != None:
                self.heartbeat_loop.stop()

    def build_droid_command(self, command_id: int, data: str) -> bytearray:
        """
        The build_droid_command function creates a bytearray that represents a command for a Droid. 
        It takes in a command_id (integer) and a data string, and returns the corresponding bytearray.

        The first byte of the bytearray represents the total length of the command in bytes. The second byte is 0x42 
        if the command id is 15, or 0x00 otherwise. The third byte is the command id itself. The fourth byte is the length 
        of the data string in bytes, plus 0x40. The remaining bytes are the data string itself, represented in hexadecimal format.

        If the data string is malformed, a ValueError is raised.

        Args:
            command_id (int): The command id to be included in the Droid command
            data (str): The data string to be included in the Droid command

        Returns:
            bytearray: The bytearray representation of the Droid command, with the given command id and data string.
        """

        data_length = len(data) // 2
        header_length = 3

        if command_id == 15:
            byte2 = 0x42
        else:
            byte2 = 0x00

        total_length = data_length + header_length
        byte1 = total_length | 0x20
        byte3 = command_id
        byte4 = data_length + 0x40

        try:
            command_bytes = bytearray([byte1, byte2, byte3, byte4])
            command_bytes.extend(bytes.fromhex(data))
        except ValueError:
            raise ValueError("Failed to pack droid command (%s) with data (%s). Data is malformed" % (command_id, data))
        
        return command_bytes

    async def send_droid_command(self, command_id: int, data: str = "") -> None:
        """
        Sends a command to the Droid, composed of a command ID and optional data.

        If the data string is malformed, a ValueError is raised.

        Args:
            command_id (int): The ID of the command to send.
            data (str): Optional data to include in the command, as a string of hexadecimal digits.
        """

        command = self.build_droid_command(command_id, data)
        logging.debug('Sending command: %s' % command.hex())
        await self.droid.write_gatt_char(DroidBluetoothCharacteristics.DroidCommandCharacteristic, bytearray.fromhex(command.hex()))

    async def send_droid_multi_command(self, command_id: int, data: str = "") -> None:
        """
        Sends a multi command to the Droid, composed of a command ID and optional data.

        If the data string is malformed, a ValueError is raised.

        Args:
            command_id (int): The ID of the command to send.
            data (str): Optional data to include in the command, as a string of hexadecimal digits.
        """

        command = "44%s%s" % ("{:02d}".format(command_id), data)
        await self.send_droid_command(DroidCommandId.MultipurposeCommand, command)

    async def get_droid_firmware_information(self) -> None:
        """
        Requests the droid firmware information. Currently the contents of this data
        is unknown. Because of this this request only returns the raw data processed by the 
        notify command processor.
        """

        await self.send_droid_command(DroidCommandId.RetrieveFirmwareInformation)
        firemware_information = await self.notify_processor.wait_for_command_response(DroidCommandId.RetrieveFirmwareInformationResponse)
        if firemware_information == None:
            raise Exception('Failed to retrieve firmware information. No response given')
        return firemware_information

    async def set_pairing_led(self, state: bool) -> None:
        """
        Sets the active state of the droid's pairing led.

        Args:
            state (bool): State to set the led to
        """

        data = "00"
        data += "ff" if state else "00"
        await self.send_droid_command(DroidCommandId.SetPairingLedState, data)

    async def set_rgb_led(self, state: bool) -> None:
        """
        Sets the active state of the droid's onboard RGB led. Currently no droids exist that use this feature
        however its added for completeness.

        Args:
            state (bool): State to set the led to
        """

        data = "00"
        data += "ff" if state else "00"
        await self.send_droid_command(DroidCommandId.SetRGBLedState, data)

    async def flash_pairing_led(self, data: str) -> None:
        """
        Flashes the droids onboard pairing led. Currently the data required for this command has not been decoded. 

        An example piece of encoded data is "020001ff01ff0aff00". This hex encoded string will flash the pairing LED 10 times at 
        a rate of once per second.
        """

        await self.send_droid_command(DroidCommandId.FlashPairingLed, data)

async def discover_droids(retry: bool = False) -> list:
    """
    Scans for nearby Bluetooth devices manufactured by Disney and have the device name of "DROID" if any are found they will be
    converted to a DroidConnection and added to a list to return. If retry is False, the function will out after a set
    period of time and return without discovering any droids.

    Args:
        retry (bool): whether or not to continue scanning until a device is found or the function is interrupted

    Returns:
        a list of DroidConnection objects representing the discovered "DROID" Bluetooth devices if any. Otherwise an empty list
    """

    async with BleakScanner() as scanner:      
        droid_connections = []
        droids = []
        
        while True:
            possible_droids = scanner.discovered_devices_and_advertisement_data
            if len(possible_droids) == 0:
                await asyncio.sleep(5)
                continue

            for possible_droid_address in possible_droids:
                ble_device, advertising_data = possible_droids[possible_droid_address]
                manufacturer_ids = list(advertising_data.manufacturer_data.keys()) if advertising_data.manufacturer_data != None else []

                if ble_device.name == "DROID" and DisneyBLEManufacturerId.DroidManufacturerId in manufacturer_ids:
                    droids.append((ble_device, advertising_data.manufacturer_data))
                    
            if len(droids) == 0:
                if not retry:
                    logging.error("Droid discovery failed. Retrying...")
                    await asyncio.sleep(5)
                    continue
                else:
                    logging.warning("Droid discovery failed. Retrying...")
                    await asyncio.sleep(5)
                    continue
            else:
                for discovered_droid in droids:
                    logging.info(f"Droid successfully discovered: [ {discovered_droid[0]} ]")
                    droid_connections.append(DroidConnection(*discovered_droid))
                break
    
    return droid_connections

async def discover_droid(retry: bool = False) -> DroidConnection:
    """
    Scans for nearby Bluetooth devices manufactured by Disney and have the device name of "DROID" and returns if found. If retry is True, the function will
    continue scanning until it finds a device or is interrupted. If retry is False, the function will time out after a
    set period of time and return without discovering a device.

    Args:
        retry (bool): whether or not to continue scanning until a device is found or the function is interrupted

    Returns:
        a DroidConnection object representing the discovered "DROID" Bluetooth device if any. Otherwise None
    """

    discovered_droids = await discover_droids(retry)
    return None if len(discovered_droids) == 0 else discovered_droids[0]


"""
Copyright (c) Jordan Maxwell, All Rights Reserved.
See LICENSE file in the project root for full license information.

This module defines the DroidCommandId and DroidMultipurposeCommand classes used to 
define the constant command identifiers to communicate with a SWGE droid.
"""

from enum import IntEnum

class DisneyBLEManufacturerId(object):
    """
    Constants representing the types of BLE beacons found in Disney parks.
    """

    DisneyiBeacon = 76
    DroidManufacturerId = 387

class DroidBluetoothCharacteristics(object):
    """
    a constants class representing the characteristics available on droid depot droids
    """

    DroidCommandCharacteristic = '09b600b1-3e42-41fc-b474-e9c0c8f0c801'
    DroidNotifyCharacteristic = '09b600b0-3e42-41fc-b474-e9c0c8f0c801'

class DroidCommandId(IntEnum):
    """
    A enum representing the available droid commands.

    Constants:
        RetrieveFirmwareInformation (int): Command to retrieve firmware information.
        SetPairingLedState (int): Command to set the pairing LED state.
        SetRGBLedState (int): Command to set the RGB LED state.
        FlashPairingLed (int): Command to flash the pairing LED.
        SetMotorSpeed (int): Command to set the motor speed.
        UnusedCommand (int): An unused command.
        ScriptActionComand (int): Command to execute a script.
        ScriptDelay (int): Command to delay script execution.
        MultipurposeCommand (int): Command to send a multipurpose command.
    """

    RetrieveFirmwareInformation = 1
    SetPairingLedState = 2
    SetRGBLedState = 3
    FlashPairingLed = 4
    SetMotorSpeed = 5
    ScriptWriteCommand = 6
    ScriptActionComand = 12
    ScriptDelay = 13
    ConnectionHeartbeat = 14
    MultipurposeCommand = 15

    RUnitHeadEvent = 128
    RetrieveFirmwareInformationResponse = 129

    @classmethod
    def valid_command(cls, value: int) -> bool:
        """
        Checks if the command id is valid
        """

        return value in cls._value2member_map_ 

class DroidMultipurposeCommand(object):
    """
    A class representing the available multipurpose commands.

    Constants:
        AudioControllerCommand (int): Command to send an audio control command.
        CenterRUnitHead (int): Command to center the R-unit head.
        RotateRUnitHead (int): Command to rotate the R-unit head.
        RotateRUnitHeadWithoutRamp (int): Command to rotate the R-unit head without ramp.
        RotateBUnitHead (int): Command to rotate the B-unit head.
        DriveBUnit (int): Command to drive the B-unit.
    """

    AudioControllerCommand = 0
    CenterRUnitHead = 1
    RotateRUnitHead = 2
    RotateRUnitHeadWithoutRamp = 3
    RotateBUnitHead = 4
    DriveBUnit = 5

class DroidAffiliation(object):
    """
    Represents a droid's affiliation for BLE interactions and audio playback.
    """

    Scoundrel = 1
    Resistenace = 5
    FirstOrder = 9


"""
Copyright (c) Jordan Maxwell, All Rights Reserved.
See LICENSE file in the project root for full license information.

This module provides a voice controller for SWGE droids that attempts to match the droid's configured affiliation
to a tone of voice based on the available audio banks in the droid's memory. 

It contains a `DroidVoiceController` class that has methods to get voice bank IDs for different tones of 
voice based on the droid's affiliation. 

The `DroidVoiceTone` class defines identifiers for the possible tones of voice to use when speaking. 
The `talk_with_animation` method of the `DroidVoiceController` class sends a command to the droid to speak with a random animation and audio file
based around the tone supplied when invoked.
"""

from droiddepot.hardware import DroidAffiliation, DroidAudioBankIdentifier
import random

class DroidVoiceTone(object):
    """
    Identifer for the possible tones of voice to use when speaking
    """

    Friendly = 1
    Indifferent = 2
    Upset = 3

class DroidVoiceController(object):
    """
    Experimental voice controller for SWGE droids that attempts to match the droid's configured affiliation
    to a tone of voice based on the available audio banks in the droid's memory.
    """

    def __init__(self, droid: object) -> None:
        self.droid = droid

    def get_friendly_voice_bank_id(self) -> int:
        """
        Returns a voice bank id for the droid's configuration that can be viewed as "happy" or "excited"
        """

        if self.droid.affiliation_id == DroidAffiliation.Resistenace:
            return DroidAudioBankIdentifier.ResistenceAudioBank
        elif self.droid.affiliation_id == DroidAffiliation.FirstOrder:
            return DroidAudioBankIdentifier.FirstOrderAudioBank
        else:
            return self.get_random_indifferent_voice_bank_id()

    def get_upset_voice_bank_id(self) -> int:
        """
        Returns a voice bank id for the droids configuration that can be viewed as "upset" or "scared"
        """

        if self.droid.affiliation_id == DroidAffiliation.Resistenace:
            return DroidAudioBankIdentifier.FirstOrderAudioBank
        elif self.droid.affiliation_id == DroidAffiliation.FirstOrder:
            return DroidAudioBankIdentifier.ResistenceAudioBank
        else:
            return self.get_random_indifferent_voice_bank_id()
        
    def get_random_indifferent_voice_bank_id(self) -> int:
        """
        Returns a random voice bank that can be viewed as "neutral" or indifferent
        """

        options = list(DroidAudioBankIdentifier.TalkingBanks)
        if self.droid.affiliation_id != DroidAffiliation.Scoundrel:
            options.remove(self.get_upset_voice_bank_id())
        return random.choice(options)

    async def talk_with_animation(self, tone: int) -> None:
        """
        Sends a command to the droid to speak with a random animation and audio file
        based around the tone supplied when invoked.
        """

        voice_script = 0
        if tone == DroidVoiceTone.Friendly:
            voice_script = self.get_friendly_voice_bank_id()
        elif tone == DroidVoiceTone.Upset:
            voice_script = self.get_upset_voice_bank_id()
        else:
            voice_script = self.get_random_indifferent_voice_bank_id()

        await self.droid.script_engine.execute_script(voice_script)


"""
Copyright (c) Jordan Maxwell, All Rights Reserved.
See LICENSE file in the project root for full license information.
"""

from droiddepot.utils import *

class OfficialDroidBeaconLocations(object):
    """
    Constants representing every official Walt Disney World and DisneyLand SWGE Droid Beacon
    """
 
    DL_Marketplace =           '0A040102A601'
    DL_BehindDroidDepot =      '0A040202A601'
    DL_Resistence =            '0A040302A601'
    DL_FirstOrder =            '0A040702A601'
    DL_DroidDepot =            '0A040318BA01'
    DL_InFrontOfOgas =         '0A0405FFA601'
    DL_MarketplaceEntertance = '0A040502A601'

    WDW_OutdoorsArea =         '0A040102A601'
    WDW_BehindDroidDepot =     '0A040202A601'
    WDW_Resistence =           '0A040302A601'
    WDW_DokOndars =            '0A040602A601'
    WDW_FirstOrder =           '0A040702A601'
    WDW_Marketplace =          '0A040618BA01'
    WDW_DroidDetector =        '0A0405FFA601'
    WDW_InFrontOfOgas =        '0A0407FFA601'



"""
Copyright (c) Jordan Maxwell, All Rights Reserved.
See LICENSE file in the project root for full license information.

This modules defines classes and helper functions for working with SWGE droid hardware. 
"""

DroidFirmwareVersion = '4b1001444411110100000000'

class DroidAudioBankIdentifier(object):
    """
    A collection of identifiers used to represent available audio banks of a droid or its connected personality chip. 
    These are used in audio playback.
    """

    GeneralUseAudioBank = 1
    DroidDepotAudioBank = 2
    ResistenceAudioBank = 3
    UnknownAudioBank = 4
    DroidDetectorAudioBank = 5
    DokOndarsAudioBank = 6
    FirstOrderAudioBank = 7
    InitialActivationAudioBank = 8
    MotorSoundAudioBank = 9
    EmptyAudioBank = 10
    BlasterAcessoryAudioBank = 11
    ThrusterAccessoryAudioBank = 12

    TalkingBanks = [
        DroidDepotAudioBank,
        ResistenceAudioBank,
        UnknownAudioBank,
        DokOndarsAudioBank,
        FirstOrderAudioBank]

class DroidAffiliation(object):
    """
    Represents a droid's affiliation for BLE interactions and audio playback.
    """

    Scoundrel = 1
    Resistenace = 5
    FirstOrder = 9

class DroidPersonalityIdentifier(object):
    """
    A collection of identifiers used to represent various types of personality chips for audio and
    affiliation lookup.

    Attributes:
        Blue (int): Identifier for the blue personality chip.
        Gray (int): Identifier for the gray personality chip.
        Red (int): Identifier for the red personality chip.
        Orange (int): Identifier for the orange personality chip.
        Purple (int): Identifier for the purple personality chip.
        Black (int): Identifier for the black personality chip.
        Red2 (int): Identifier for the Red 2 personality chip. Same value as CB23.
        Yellow (int): Identifier for the yellow personality chip.
        DarkBlue (int): Identifier for the dark blue personality chip.

        C110P (int): Identifier for the C1-10P droid.
        CB23 (int): Identifier for the CB-23 droid.
        BUnit (int): Identifier for the BD unit droid.
        RUnit (int): Identifier for the R unit droid.
        BBUnit (int): Identifier for the BB unit droid.

        ChipAudioCount (dict): Dictionary containing a total count of available audio by audio bank id and by droid/personality chip
    """

    # Personality Chips
    Blue = 3
    Gray = 4
    Red = 5
    Orange = 6
    Purple = 7
    Black = 8
    Yellow = 10
    Red2 = 11
    DarkBlue = 13

    # Droids
    BUnit = 1
    RUnit = 2
    CB23 = 9
    C110P = 11
    DO = 12
    BBUnit = 14  

    ChipAudioCount = {
        DroidAudioBankIdentifier.GeneralUseAudioBank:           { Blue: 5, Gray: 4, Red: 5, Orange: 5, Purple: 4, Black: 3, Red2: 5, CB23: 5, C110P: 6,  Yellow: 4, DarkBlue: 4, BUnit: 5, RUnit: 4, BBUnit: 5 },
        DroidAudioBankIdentifier.DroidDepotAudioBank:           { Blue: 5, Gray: 5, Red: 5, Orange: 5, Purple: 4, Black: 6, Red2: 5, CB23: 5, C110P: 13, Yellow: 5, DarkBlue: 4, BUnit: 6, RUnit: 4, BBUnit: 3 },
        DroidAudioBankIdentifier.ResistenceAudioBank:           { Blue: 5, Gray: 5, Red: 5, Orange: 5, Purple: 5, Black: 5, Red2: 5, CB23: 5, C110P: 5,  Yellow: 5, DarkBlue: 4, BUnit: 6, RUnit: 3, BBUnit: 3 },
        DroidAudioBankIdentifier.UnknownAudioBank:              { Blue: 1, Gray: 1, Red: 1, Orange: 1, Purple: 1, Black: 1, Red2: 1, CB23: 1, C110P: 1,  Yellow: 1, DarkBlue: 1, BUnit: 1, RUnit: 1, BBUnit: 1 },
        DroidAudioBankIdentifier.DroidDetectorAudioBank:        { Blue: 1, Gray: 1, Red: 1, Orange: 1, Purple: 1, Black: 1, Red2: 1, CB23: 1, C110P: 1,  Yellow: 1, DarkBlue: 1, BUnit: 1, RUnit: 1, BBUnit: 1 },
        DroidAudioBankIdentifier.DokOndarsAudioBank:            { Blue: 5, Gray: 3, Red: 3, Orange: 3, Purple: 4, Black: 5, Red2: 5, CB23: 5, C110P: 6,  Yellow: 5, DarkBlue: 4, BUnit: 5, RUnit: 4, BBUnit: 5 },
        DroidAudioBankIdentifier.FirstOrderAudioBank:           { Blue: 3, Gray: 5, Red: 3, Orange: 3, Purple: 5, Black: 3, Red2: 5, CB23: 5, C110P: 6,  Yellow: 5, DarkBlue: 5, BUnit: 5, RUnit: 5, BBUnit: 5 },
        DroidAudioBankIdentifier.InitialActivationAudioBank:    { Blue: 1, Gray: 1, Red: 1, Orange: 1, Purple: 1, Black: 1, Red2: 1, CB23: 1, C110P: 1,  Yellow: 1, DarkBlue: 1, BUnit: 1, RUnit: 1, BBUnit: 1 },
        DroidAudioBankIdentifier.MotorSoundAudioBank:           { Blue: 1, Gray: 1, Red: 1, Orange: 1, Purple: 1, Black: 1, Red2: 1, CB23: 1, C110P: 1,  Yellow: 1, DarkBlue: 1, BUnit: 0, RUnit: 1, BBUnit: 0 },
        DroidAudioBankIdentifier.EmptyAudioBank:                { Blue: 0, Gray: 0, Red: 0, Orange: 0, Purple: 0, Black: 0, Red2: 0, CB23: 0, C110P: 0,  Yellow: 0, DarkBlue: 0, BUnit: 0, RUnit: 0, BBUnit: 0 },
        DroidAudioBankIdentifier.BlasterAcessoryAudioBank:      { Blue: 2, Gray: 2, Red: 2, Orange: 2, Purple: 2, Black: 2, Red2: 2, CB23: 2, C110P: 2,  Yellow: 2, DarkBlue: 2, BUnit: 0, RUnit: 2, BBUnit: 0 },
        DroidAudioBankIdentifier.ThrusterAccessoryAudioBank:    { Blue: 2, Gray: 2, Red: 2, Orange: 2, Purple: 2, Black: 2, Red2: 2, CB23: 2, C110P: 2,  Yellow: 2, DarkBlue: 2, BUnit: 0, RUnit: 2, BBUnit: 0 }
    }

    ChipAffiliation = {
        Blue:   DroidAffiliation.Resistenace,
        Gray:   DroidAffiliation.Scoundrel,
        Red:    DroidAffiliation.FirstOrder,
        Orange: DroidAffiliation.Resistenace,
        Purple: DroidAffiliation.Scoundrel,
        Black:  DroidAffiliation.FirstOrder,
        Red2:   DroidAffiliation.Scoundrel,
        Yellow: DroidAffiliation.Resistenace,

        CB23:   DroidAffiliation.Scoundrel,
        DO:     DroidAffiliation.Resistenace,
        C110P:  DroidAffiliation.Resistenace,
        RUnit:  DroidAffiliation.Scoundrel,
        BUnit:  DroidAffiliation.Resistenace,
        BBUnit: DroidAffiliation.Scoundrel
    }   

    ChipShutdownTrack = {
        Blue:   (DroidAudioBankIdentifier.FirstOrderAudioBank, 2),
        Gray:   (DroidAudioBankIdentifier.FirstOrderAudioBank, 4),
        Red:    (DroidAudioBankIdentifier.FirstOrderAudioBank, 2),
        Orange: (DroidAudioBankIdentifier.FirstOrderAudioBank, 3),
        Purple: (DroidAudioBankIdentifier.FirstOrderAudioBank, 1),
        Black:  (DroidAudioBankIdentifier.FirstOrderAudioBank, 1),
        Red2:   (DroidAudioBankIdentifier.FirstOrderAudioBank, 1),
        Yellow: (DroidAudioBankIdentifier.FirstOrderAudioBank, 1),

        CB23:   (DroidAudioBankIdentifier.FirstOrderAudioBank, 1),
        DO:     (DroidAudioBankIdentifier.FirstOrderAudioBank, 1),
        C110P:  (DroidAudioBankIdentifier.FirstOrderAudioBank, 1),
        RUnit:  (DroidAudioBankIdentifier.FirstOrderAudioBank, 2),
        BUnit:  (DroidAudioBankIdentifier.FirstOrderAudioBank, 1),
        BBUnit: (DroidAudioBankIdentifier.FirstOrderAudioBank, 3),
    }

class DroidLedIdentifier(object):
    """
    A collection of LED identifiers for a droid.

    Attributes:
        RUnitLeftHeadLed (int): Identifier for the left head LED on an R unit.
        RUnitMiddleHeadLed (int): Identifier for the middle head LED on an R unit.
        RUnitRightHeadLed (int): Identifier for the right head LED on an R unit.
        RUnitLeftAccessoryLed (int): Identifier for the left accessory LED on an R unit.
        RUnitRightAcessoryLed (int): Identifier for the right accessory LED on an R unit.
        BUnitHeadLeds (int): Identifier for the head LEDs on a BB unit.
        BUnitLed0Green (int): Identifier for LED 0 green on a BD unit.
        BUnitLED0Green (int): Identifier for LED 0 green on a BD unit.
        BUnitLED0Red (int): Identifier for LED 0 red on a BD unit.
        BUnitLED1Blue (int): Identifier for LED 1 blue on a BD unit.
        BUnitLED1Green (int): Identifier for LED 1 green on a BD unit.
        BUnitLED1Red (int): Identifier for LED 1 red on a BD unit.
        BUnitLED2Blue (int): Identifier for LED 2 blue on a BD unit.
        BUnitLED2Green (int): Identifier for LED 2 green on a BD unit.
        BUnitLED2Red (int): Identifier for LED 2 red on a BD unit.
        BUnitLED3Blue (int): Identifier for LED 3 blue on a BD unit.
        BUnitLED3Green (int): Identifier for LED 3 green on a BD unit.
        BUnitLED3Blue (int): Identifier for LED 3 blue on a BD unit.
        BUnitLeftEyeLed (int): Identifier for the left eye LED on a BD unit.
        BUnitRightEyeLed (int): Identifier for the right eye LED on a BD unit.
    """

    # R and C units
    RUnitLeftHeadLed = 1
    RUnitMiddleHeadLed = 2
    RUnitRightHeadLed = 4
    RUnitLeftAccessoryLed = 8
    RUnitRightAcessoryLed = 16

    # BB Units
    BUnitHeadLeds = 1

    # BD Units
    BUnitLed0Green = 0
    BUnitLED0Green = 1
    BUnitLED0Red = 2
    BUnitLED1Blue = 3
    BUnitLED1Green = 4
    BUnitLED1Red = 5
    BUnitLED2Blue = 6
    BUnitLED2Green = 7
    BUnitLED2Red = 8
    BUnitLED3Blue = 9
    BUnitLED3Green = 10
    BUnitLED3Blue = 11
    BUnitLeftEyeLed = 12
    BUnitRightEyeLed = 13

def get_personality_id(droid_id: int, chip_id: int = None) -> int:
    """
    Returns the proper personality id given a droid's identifier and its current personality chip if any.

    Args:
        droid_id (int): The identifier representing the droid from the DroidPersonalityIdentifier class
        chip_id (int): The optional chip identifier representing the droid's personality chip from the DroidPersonalityIdentifier class.

    Returns:
        an integer representing the droid's personality that can be used for audio queries.
    """

    return droid_id if chip_id == None else chip_id

def get_available_audio_in_bank(bank_id: int, personality_id: int) -> int:
    """
    Returns the amount of audio clips that are available to play in a given audio bank depending on the droid/personality chip

    Args:
        bank_id (int): Identifier representing the audio bank to search
        personality_id (int): Represents the droid's current configured personality.

    Returns:
        an integer representing the total number of available audio clips
    """

    if bank_id not in DroidPersonalityIdentifier.ChipAudioCount:
        return 0
    
    bank_info = DroidPersonalityIdentifier.ChipAudioCount[personality_id]
    return 0 if personality_id not in bank_info else bank_info[personality_id]

def get_shutdown_audio_track(affiliation_id: int) -> tuple:
    """
    Returns a tuple containing the audio bank and sounud id that should be played when the droid goes to sleep
    based on its afilliation.

    Args:
        affiliation_id (int): The identnfier that represents the droid's affiliation

    Returns:
        A tuple(bank_id, sound_id) containing the droid's shutdown audio information. If the information cannot be found the
        default is returned of (7, 1).
    """

    if affiliation_id not in DroidPersonalityIdentifier.ChipShutdownTrack:
        return (DroidAudioBankIdentifier.FirstOrderAudioBank, 1)
    
    return DroidPersonalityIdentifier.ChipShutdownTrack[affiliation_id]

def get_personality_affiliation(personality_id: int) -> int:
    """
    """

    if personality_id not in DroidPersonalityIdentifier.ChipAffiliation:
        return DroidAffiliation.Scoundrel
    
    return DroidPersonalityIdentifier.ChipAffiliation[personality_id]


"""
Copyright (c) Jordan Maxwell, All Rights Reserved.
See LICENSE file in the project root for full license information.

Utility methods module for PyDroidDepot. 
"""

def int_to_hex(num: int) -> str:
    """
    Converts an integer to a hexadecimal string.

    Args:
        num (int): The integer to be converted to a hexadecimal string.

    Returns:
        str: The hexadecimal string representation of the integer.
    """

    hex_str = hex(num)[2:]  # Get the hex string without the '0x' prefix
    if len(hex_str) % 2 != 0:
        hex_str = "0" + hex_str
    
    return hex_str

def hex_to_int(hex_str: str) -> int:
    """
    Converts a hexadecimal string to an integer.

    Args:
        hex_str (str): The hexadecimal string to be converted to an integer.

    Returns:
        int: The integer representation of the hexadecimal string.
    """
    
    return int(hex_str, 16)

def dbm_to_hex(dbm_val: int) -> str:
    """
    Convert a dBm value to its corresponding hex integer string.

    Args:
        dbm_val (int): The dBm value to convert.

    Returns:
        str: The hex integer string corresponding to the input dBm value.
    """
        
    dbm_val = int((dbm_val - 0x80) / -1)
    return int_to_hex(dbm_val)

def hex_to_dbm(hex_str: str) -> float:
    """
    Convert a hex integer string to its corresponding dBm value.

    Args:
        hex_str (str): The hex integer string to convert.

    Returns:
        float: The dBm value corresponding to the input hex integer string.
    """

    # Convert the hex integer string to an integer value
    hex_int = int(hex_str, 16)

    # Calculate the dBm value based on the provided range
    dbm_val = (hex_int - 0x80) * -1

    return dbm_val


"""
Copyright (c) Jordan Maxwell, All Rights Reserved.
See LICENSE file in the project root for full license information.

This module defines classes for controlling audio and LEDs for a droid. It contains three classes:
1. DroidAudioCommand: A collection of audio commands for a droid
2. DroidLedIdentifier: A collection of LED identifiers for a droid
3. DroidAudioController: Represents an audio controller for a Droid and has methods for controlling audio and LEDs
"""

from enum import IntEnum
from droiddepot.utils import  int_to_hex
from droiddepot.protocol import DroidMultipurposeCommand, DroidAffiliation
from droiddepot.hardware import DroidLedIdentifier, get_shutdown_audio_track

class DroidAudioCommand(IntEnum):
    """
    A collection of audio commands for a droid.

    Attributes:
        RetrieveDroidType (int): Command to retrieve the droid type.
        RetrievePersonalityChip (int): Command to retrieve the personality chip.
        RetrieveAffiliation (int): Command to retrieve the affiliation.
        SetVolume (int): Command to set the audio volume.
        UnknownCommand1 (int): Unknown audio command.
        PlayAudioFromGroupByValue (int): Command to play audio from a group by value.
        PlayAudioFromGroupByValueWithoutLeds (int): Command to play audio from a group by value without LEDs.
        PlayAudioFromSelectedGroup (int): Command to play audio from the selected group.
        CycleAudioFromSelectedGroup (int): Command to cycle audio from the selected group.
        SetSelectedSoundBank (int): Command to set the selected sound bank.
        SetLoopedAudio (int): Command to set the audio to loop.
        UnknownCommand2 (int): Unknown audio command.
        UnknownCommand3 (int): Unknown audio command.
        FlashHeadLeds (int): Command to flash the head LEDs.
        SetLedOn (int): Command to set a specific LED on.
        SetLedOff (int): Command to set a specific LED off.
        DisableHeadLeds (int): Command to disable the head LEDs.
        EnableHeadLeds (int): Command to enable the head LEDs.
    """

    RetrieveDroidType = 1
    RetrievePersonalityChip = 8
    RetrieveAffiliation = 10
    SetVolume = 14
    UnknownCommand1 = 15
    PlayAudioFromGroupByValue = 16
    PlayAudioFromGroupByValueWithoutLeds = 17
    PlayAudioFromSelectedGroup = 24
    CycleAudioFromSelectedGroup = 28
    SetSelectedSoundBank = 31
    LoopSoundBank = 33
    UnknownCommand2 = 66
    UnknownCommand3 = 68
    FlashHeadLeds = 69
    SetLedOn = 72
    SetLedOff = 73
    DisableHeadLeds = 74
    EnableHeadLeds = 75

class DroidLedIdentifier(object):
    """
    Constants relating to various Leds found in droid depot droids.
    """

    # BD Head Leds. These Leds are RGB. 
    # To change the RGB values set the values as follows base (blue), base + 1 (green), base + 2 (red)
    BDUnitLedZero = 0
    BDUnitLedOne = 3
    BDUnitLedTwo = 6
    BDUnitLedThree = 9 
    BDUnitLeftEye = 12
    BDUnitRightEye = 13

    RUnitLeftHeadLed = 1
    RUnitMiddleHeadLed = 2
    RUnitRightHeadLed = 4
    RUnitLeftAccessory = 8
    RUnitRightAccessory = 16

    BBUnitHeadLed = 1

class DroidAudioController(object):
    """
    Represents an audio controller for a Droid.

    Args:
        droid (DroidConnection): The DroidConnection this audio controller is for.
    """

    def __init__(self, droid: object) -> None:
        """
        Initializes a new instance of the DroidAudioController class.

        Args:
            droid (DroidConnection): The DroidConnection this audio controller is for.
        """

        self.droid = droid
        self.sound_bank = 0
        self.disabled_leds = []
        self.turned_on_leds = []

    async def execute_audio_command(self, command_id: int, data: str = "00") -> None:
        """
        Executes an audio command on the Droid.

        Args:
            command_id (int): The ID of the audio command to execute.
            data (str): The data to send with the audio command, if any.

        Returns:
            None
        """

        command_id = int_to_hex(command_id)
        command_data = "%s%s"  % (command_id, data)
        await self.droid.send_droid_multi_command(DroidMultipurposeCommand.AudioControllerCommand, command_data)

    async def play_audio(self, sound_id: int = None, bank_id: int = None, cycle: bool = False, volume: int = None) -> None:
        """
        Plays audio on the Droid.

        Args:
            sound_id (int): The ID of the sound to play. Defaults to 0 if not provided.
            bank_id (int): The ID of the audio bank to use. Defaults to 0 if not provided.
            cycle (bool): If True, cycles through the audio files in the selected audio bank. Defaults to False.
            volume (int): The volume to play the audio at, in the range [0, 100]. If not provided, the Droid's current volume is used.

        Returns:
            None
        """

        if volume:
            await self.set_volume(volume)

        bank_id = bank_id - 1 if bank_id != None else 0
        if bank_id and (not hasattr(self, "sound_bank") or self.sound_bank != bank_id):
            await self.set_audio_bank(bank_id)

        sound_id = int_to_hex(sound_id - 1 if sound_id != None else 0)
        bank_id = int_to_hex(bank_id)

        audio_command = "00"
        audio_parameter = "00"

        if sound_id:
            audio_command = DroidAudioCommand.PlayAudioFromSelectedGroup
            audio_parameter = sound_id
        elif cycle:
            audio_command = DroidAudioCommand.CycleAudioFromSelectedGroup
        else:
            audio_command = DroidAudioCommand.PlayAudioFromGroupByValue
            audio_parameter = bank_id
        
        await self.execute_audio_command(audio_command, audio_parameter)

    async def play_shutdown_audio(self) -> None:
        """
        Plays the droid's shutdown audio based on its configured personality id
        """

        bank_id, sound_id = get_shutdown_audio_track(self.droid.personality_id)
        await self.play_audio(sound_id=sound_id, bank_id=bank_id, cycle=True)

    async def set_audio_bank(self, bank_id: int) -> None:
        """
        Sets the selected audio bank on the Droid.

        Args:
            bank_id (int): The ID of the audio bank to select.
        """

        bank_id = int_to_hex(bank_id if bank_id != None else 0)
        self.sound_bank = bank_id

        await self.execute_audio_command(DroidAudioCommand.SetSelectedSoundBank, bank_id)

    async def set_volume(self, volume_level: int) -> None:
        """
        Sets the volume of the audio playback on the Droid.

        Args:
            volume_level (int): The volume level to set.
        """

        volume_level = int_to_hex(volume_level if volume_level != None else 0)
        await self.execute_audio_command(DroidAudioCommand.SetVolume, volume_level)

    async def reset_head_leds(self) -> None:
        """
        """

        await self.enable_head_led(31)

    async def disable_head_led(self, led_identifier: int) -> None:
        """
        """

        await self.execute_audio_command(DroidAudioCommand.DisableHeadLeds, led_identifier)

        if led_identifier not in self.disabled_leds:
            self.disabled_leds.append(led_identifier)

    async def enable_head_led(self, led_identifier: int) -> None:
        """
        """

        led_identifier = int_to_hex(led_identifier)
        await self.execute_audio_command(DroidAudioCommand.EnableHeadLeds, led_identifier)

        if led_identifier in self.disabled_leds:
            self.disabled_leds.remove(led_identifier)

    async def turn_on_led(self, led_identifier: int) -> None:
        """
        """

        led_identifier = int_to_hex(led_identifier)
        await self.execute_audio_command(DroidAudioCommand.SetLedOn, led_identifier)

        if not led_identifier in self.turned_on_leds:
            self.turned_on_leds.append(led_identifier)

    async def turn_off_led(self, led_identifier: int) -> None:
        """
        """

        led_identifier = int_to_hex(led_identifier)
        await self.execute_audio_command(DroidAudioCommand.SetLedOff, led_identifier)

        if led_identifier in self.turned_on_leds:
            self.turned_on_leds.remove(led_identifier)


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from pathlib import Path
repository_directory = Path(__file__).parent
long_description = (repository_directory / "README.md").read_text()

setup(
    name='pyDroidDepot',
    description="Module for controlling droids built and purchased at the Droid Depot in Disney's Galaxys Edge",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    version='1.1.0',
    author='Jordan Maxwell',
    maintainer='Jordan Maxwell',
    url='https://github.com/thetestgame/pyDroidDepot',
    packages=['droiddepot'],
    classifiers=[
        'Programming Language :: Python :: 3',
    ])


"""
Copyright (c) Jordan Maxwell, All Rights Reserved.
See LICENSE file in the project root for full license information.

This module provides classes for controlling the motor functions of a SWGE DroidDepot droid.
"""

from enum import IntEnum
from droiddepot.utils import int_to_hex
from droiddepot.protocol import DroidCommandId, DroidMultipurposeCommand

class DroidMotorDirection(object):
    """
    Enumeration of motor directions.
    """

    Forward = 0
    Backwards = 8

    Left = 0
    Right = 8

class DroidMotorIdentifier(IntEnum):
    """
    Enumeration of motor identifiers.
    """

    LeftMotor = 0
    RightMotor = 1
    HeadMotor = 2

class DroidMotorEvent(IntEnum):
    """
    """

    MotorStarted = 2
    MotorMovingRight = 3
    MotorMovingLeft = 4
    MotorStopped = 130
    MotorHitLeftLimit = 131
    MotorHitRightLimit = 132

class DroidMotorController(object):
    """
    Class for controlling the motor functions of a SWGE DroidDepot droid.
    """

    def __init__(self, droid: object) -> None:
        """
        Initializes a new instance of the DroidMotorController class.

        Args:
            droid: A DroidConnection object representing the connection to the droid.
        """

        self.droid = droid
        self.__motor_event_handlers = []

    def subscribe_runit_head_motor_events(self, handler: object) -> None:
        """
        """

        if handler not in self.__motor_event_handlers:
            self.__motor_event_handlers.append(handler)

    def unsubscribe_runit_head_motor_events(self, handler: object) -> None:
        """

        """

        if handler in self.__motor_event_handlers:
            self.__motor_event_handlers.remove(handler)

    async def process_runit_head_motor_event(self, event_id: int) -> None:
        """
        """

        for handler in self.__motor_event_handlers:
            handler(DroidMotorEvent(event_id))
    
    async def send_motor_speed_command(self, direction: int, motor_id: int, speed: int = 160, ramp_speed: int = 300, delay = 0) -> None:
        """
        Sends a motor speed command to the droid.

        Args:
            direction (int): An integer representing the motor direction. Should be one of the values defined in the DroidMotorDirection class.
            motor_id (int): An integer representing the motor identifier. Should be one of the values defined in the DroidMotorIdentifier class.
            speed (int): An integer representing the motor speed. Defaults to 160.
            ramp_speed (int): An integer representing the motor ramp speed. Defaults to 300.
        """

        delay_hex = int_to_hex(delay)
        if len(delay_hex) < 4:
            missing = 4 - len(delay_hex)
            for x in range(missing):
                delay_hex = '0' + delay_hex

        motor_select = "%s%d" % (direction, motor_id)
        motor_command = "%s%s%s%s" % (motor_select, int_to_hex(speed), int_to_hex(ramp_speed), delay_hex)
        await self.droid.send_droid_command(DroidCommandId.SetMotorSpeed, motor_command)

    async def set_movement_speed(self, direction: int, speed: int = 100, ramp_speed: int = 300) -> None:
        """
        Sends a motor speed command to the droid to both the left and right movement motors.

        Args:
            direction (int): An integer representing the motor direction. Should be one of the values defined in the DroidMotorDirection class.
            speed (int): An integer representing the motor speed. Defaults to 160.
            ramp_speed (int): An integer representing the motor ramp speed. Defaults to 300.
        """

        await self.send_motor_speed_command(direction, DroidMotorIdentifier.LeftMotor, speed, ramp_speed)
        await self.send_motor_speed_command(direction, DroidMotorIdentifier.RightMotor, speed, ramp_speed)

    async def rotate_head(self, direction: int, speed: int = 160, ramp_speed: int = 300) -> None:
        """
        Rotates the head of the droid.

        Args:
            direction (int): An integer representing the direction to rotate the head. Should be one of the values defined in the DroidMotorDirection class.
            speed (int): An integer representing the rotation speed. Defaults to 160.
            ramp_speed (int): An integer representing the rotation ramp speed. Defaults to 300.
        """

        if direction != DroidMotorDirection.Left and direction != DroidMotorDirection.Right:
            raise ValueError("Direction is invalid. Expected values are 0 (Left) and 8 (Right)")
        
        await self.send_motor_speed_command(direction, DroidMotorIdentifier.HeadMotor, speed, ramp_speed)

    async def stop_head(self) -> None:
        """
        Stops the rotation of the head.
        """

        await self.rotate_head(DroidMotorDirection.Left, 0)

    async def center_head(self, speed: int = 255, offset: int  = 0) -> None:
        """
        Centers the head of the droid.

        Args:
            speed (int): An integer representing the speed at which to center the head. Defaults to 255.
            offset (int): An integer representing the offset from center. Defaults to 0.
        """

        command_data = "%s%s" % (int_to_hex(speed), int_to_hex(offset))
        await self.droid.send_droid_multi_command(DroidMultipurposeCommand.CenterRUnitHead, command_data)
